# -*- coding: utf-8 -*-

#  Copyright © 2013-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
import shlex

from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gio
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import PeasGtk
from gi.repository import Vte
from gi.repository import GtkSource
from gi.repository import Pango
from gi.repository import Gedit

from .helper import ConnectHelper
from .debugger import debuggers

DATA_DIR = os.path.dirname(__file__)


CATEGORY_BREAKPOINT = 'debug-break'
    
    
class ViewActivatable(GObject.Object, Gedit.ViewActivatable, ConnectHelper):
    __gtype_name__ = "GeditDebuggerView"
    view = GObject.property(type=Gedit.View)
    frame_bg_color = Gdk.RGBA()
    frame_bg_color.parse('#e6c0ac')
    
    def __init__(self):
        GObject.Object.__init__(self)
        ConnectHelper.__init__(self)
        
    def do_activate(self):
        def set_iconfile(attributes, filename):
            iconfile = Gio.File.new_for_path(os.path.join(DATA_DIR, 'images', filename))
            attributes.set_gicon(Gio.FileIcon.new(iconfile))
        attributes = GtkSource.MarkAttributes()
        set_iconfile(attributes, 'anjuta-pcmark-16.png')
        self.view.set_mark_attributes('debug-current-line', attributes, 2)
        
        attributes = GtkSource.MarkAttributes()
        attributes.set_background(self.frame_bg_color)
        self.view.set_mark_attributes('debug-frame', attributes, 1)
        
        attributes = GtkSource.MarkAttributes()
        set_iconfile(attributes, 'anjuta-breakpoint-enabled-16.png')
        self.view.set_mark_attributes(CATEGORY_BREAKPOINT, attributes, 1)
        
    def do_deactivate(self):
        attributes = GtkSource.MarkAttributes()
        self.view.set_mark_attributes('debug-current-line', attributes, 1)
        attributes = GtkSource.MarkAttributes()
        self.view.set_mark_attributes('debug-frame', attributes, 1)
        attributes = GtkSource.MarkAttributes()
        self.view.set_mark_attributes(CATEGORY_BREAKPOINT, attributes, 1)
        
        
class Breakpoints (Gtk.ListStore):
    # list of [cnum, snum, GtkSourceMark, filename, lineno]
    def __init__(self):
        Gtk.ListStore.__init__(self, int, int, object, str, int)
        self.next_cnum = 1
        self.sort_model = None
        
    def child_path(self, tpath):
        if isinstance(tpath, str):
            tpath = Gtk.TreePath.new_from_string(tpath)
        return self.sort_model.convert_path_to_child_path(tpath)
        
    def fileinfo(self, tpath):
        return self[tpath][3:5]
        
    def identifiers(self, tpath):
        return self[tpath][1:3]
        
    def new_sort_model(self):
        model = Gtk.TreeModel.sort_new_with_model(self)
        model.set_sort_column_id(3, Gtk.SortType.ASCENDING)
        def sortfunc(model, a, b, unused_data):
            a = model.fileinfo(a)
            b = model.fileinfo(b)
            if a < b:   return -1
            if a > b:   return 1
            return 0
        model.set_sort_func(3, sortfunc)
        self.sort_model = model
        return model
        
    def clear(self):
        Gtk.ListStore.clear(self)
        self.next_cnum = 1
        
    def lines_in_file(self, filename):
        return ((row, row[4]) for row in self if row[3] == filename)
        
    def add(self, sourcemark, filename, lineno):
        cnum = self.next_cnum
        self.append([cnum, 0, sourcemark, filename, lineno])
        self.next_cnum += 1
        return cnum
        
    def remove_by_sourcemark(self, sourcemark):
        for i, (cn, sn, mk, fn, ln) in enumerate(self):
            if sourcemark == mk:
                Gtk.ListStore.__delitem__(self, i)
                return sn
        return None
        
    def update(self, breakpoint):
        cnum, snum, filename, lineno = breakpoint
        for i, (cn, sn, mk, fn, ln) in enumerate(self):
            if cnum == cn:
                if snum == 0:  # requested breakpoint was invalid
                    del self[i]
                    return mk, fn, ln, 0
                else:  # requested breakpoint confirmed, maybe lineno changed
                    self[i] = [cn, snum, mk, filename, lineno]
                    return mk, fn, ln, lineno
        # requested and confirmed breakpoint does not exist any more, ignore it
        return None, filename, 0, 0
        
    @staticmethod
    def update_sourcemark(row, sourcemark):
        row[2] = sourcemark
        
        
class WindowActivatable(GObject.Object, Gedit.WindowActivatable, ConnectHelper, PeasGtk.Configurable):
    __gtype_name__ = "GeditDebuggerWindow"
    window = GObject.property(type=Gedit.Window)
    obj = property(lambda self: self.window)
    uifile = os.path.join(DATA_DIR, 'main.ui')
    menufile = os.path.join(DATA_DIR, 'menu.ui')
    modifier_mask = Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.CONTROL_MASK
    
    def __init__(self):
        GObject.Object.__init__(self)
        ConnectHelper.__init__(self)
        self.debug_active = False  # whether the debug panels are visible
        self.debugger = None
        self.current_line = None # debugged line: (document, filename, lineno) or None
        self.files_with_frame_marks = []
        self.breakpoints = Breakpoints()
        self.frameno = 0
        self.keypath = None
        self.no_update_keypath = False
        
    def do_activate(self):
        '''called when the plugin is activated
        
        Load the ui, but don't show the panels. The debugger remains in
        deactivated state.
        '''
        builder = Gtk.Builder()
        builder.add_from_file(self.uifile)
        self.actiongroup_main = builder.get_object('actiongroup_main')
        uimanager = self.window.get_ui_manager()
        uimanager.insert_action_group(self.actiongroup_main)
        self.merge_id = uimanager.add_ui_from_file(self.menufile)
        
        def get_icon_action(actionname, filename):
            action = builder.get_object(actionname)
            iconfile = Gio.File.new_for_path(os.path.join(DATA_DIR, 'images', filename))
            action.set_gicon(Gio.FileIcon.new(iconfile))
            return action
        self.action_debug_toggle = get_icon_action('action_debug_toggle', 'bug_nicu_buculei_01.png')
        self.connectx(self.action_debug_toggle, 'activate', self.on_action_debug_toggle)
        self.actions_visible_if_active = [
            (get_icon_action('action_step', 'anjuta-step-into-24.png'), self.on_action_step_activate),
            (get_icon_action('action_next', 'anjuta-step-over-24.png'), self.on_action_next_activate),
            (get_icon_action('action_return', 'anjuta-step-out-24.png'), self.on_action_return_activate),
            (get_icon_action('action_run_to', 'anjuta-run-to-cursor-24.png'),
                                                            self.on_action_run_to_activate),
            (get_icon_action('action_jump_to', 'jump-to.png'), self.on_action_jump_to_activate),
            (builder.get_object('action_cont'), self.on_action_cont_activate),
            (builder.get_object('action_cont_nopause'), self.on_action_cont_nopause_activate),
            (builder.get_object('action_pause'), self.on_action_pause_activate),
            (get_icon_action('action_break', 'anjuta-breakpoint-toggle-24.png'),
                                                                    self.on_action_break_activate),
            (builder.get_object('action_quit'), self.on_action_quit_activate),
        ]
        
        # connect tabs, if the plugin is activated in the gedit-preferences-dialog,
        # already open tabs must be connected manually
        for view in self.window.get_views():
            doc = view.get_buffer()
            self.connectx(doc, 'loaded', self.on_document_loaded, view)
            self.connectx(view, 'line-mark-activated', self.on_view_line_mark_activated)
        self.connectx(self.window, 'tab-added', self.on_tab_added)
        #self.connectx(self.window, 'active-tab-changed', on_active_tab_changed)
        self.connectx(self.window, 'tab-removed', self.on_tab_removed)
        
        # sidepanel
        self.paned_sidepanel = builder.get_object('paned_sidepanel')
        self.treeview_stack = builder.get_object('treeview_stack')
        self.treeview_namespace = builder.get_object('treeview_namespace')
        self.connectx(self.treeview_stack, 'cursor-changed', self.on_treeview_stack_cursor_changed)
        self.connectx(self.treeview_stack, 'row-activated', self.on_treeview_stack_row_activated)
        self.connectx(self.treeview_namespace, 'cursor-changed', self.on_treeview_namespace_cursor_changed)
        self.connectx(self.treeview_namespace, 'row-activated', self.on_treeview_namespace_row_activated)
        
        # bottompanel
        self.notebook_bottompanel = builder.get_object('notebook_bottompanel')
        treeview_breakpoints = builder.get_object('treeview_breakpoints')
        treeview_breakpoints.set_model(self.breakpoints.new_sort_model())
        box_console = builder.get_object('box_console')
        self.terminal = Vte.Terminal()
        box_console.pack_end(self.terminal, True, True, 0)
        self.terminal.show()
        self.connectx(self.notebook_bottompanel, 'switch-page', self.on_notebook_bottompanel_switch_page)
        self.connectx(treeview_breakpoints, 'row-activated', self.on_treeview_breakpoints_row_activated)
        self.connectx(treeview_breakpoints, 'key-release-event',
                                                        self.on_treeview_breakpoints_key_release_event)
        
        # parameters tab
        self.grid_parameters = builder.get_object('grid_parameters')
        self.entry_program = builder.get_object('entry_program')
        self.entry_arguments = builder.get_object('entry_arguments')
        self.entry_workingdir = builder.get_object('entry_workingdir')
        button_set_program = builder.get_object('button_set_program')
        button_set_workingdir = builder.get_object('button_set_workingdir')
        self.connectx(self.entry_arguments, 'activate', self.on_entry_arguments_activate)
        self.connectx(self.entry_workingdir, 'activate', self.on_entry_workingdir_activate)
        self.connectx(button_set_program, 'clicked', self.on_button_set_program_activate)
        self.connectx(button_set_workingdir, 'clicked', self.on_button_set_workingdir_activate)
        
    def do_deactivate(self):
        '''called when the plugin is deactivated'''
        self.disconnect_all()
        self.deactivate_debugger()
        self.breakpoints.clear()
        uimanager = self.window.get_ui_manager()
        uimanager.remove_ui(self.merge_id)
        uimanager.remove_action_group(self.actiongroup_main)
        panel = self.window.get_side_panel()
        panel.remove_item(self.paned_sidepanel)
        self.terminal = None
        
    #def do_create_configure_widget(self):
    #    return Gtk.Label('Hallo Welt!\nDas hier ist der Konfigurations-Dialog')
        
    @staticmethod
    def add_debugger_panel(panel, idname, widget):
        icon = Gtk.Image.new_from_file(os.path.join(DATA_DIR, 'images', 'bug_nicu_buculei_01.png'))
        panel.add_item(widget, idname, "Debugger", icon)
        panel.activate_item(widget)
        
    def activate_debugger(self):
        '''Activate the debugger
        
        Show the bottom panel, the debugger remains in the not running state.
        '''
        if self.debug_active:
            return
        self.debug_active = True
        self.action_debug_toggle.set_active(True)
        for action, handler in self.actions_visible_if_active:
            action.set_visible(True)
            self.connectx(action, 'activate', handler)
        self.terminal.reset(True, True)
        self.terminal.feed(b'Console for Gedit Python Debugger Plugin\r\n')
        
        panel = self.window.get_bottom_panel()
        self.add_debugger_panel(panel, "DebuggerBottomPanel", self.notebook_bottompanel)
        
    def deactivate_debugger(self):
        '''Deactivate the debugger
        
        Stop debugger and remove the panels
        '''
        self.disconnect_object(self.debugger)
        self.end_debugging()
        if not self.debug_active:
            return
            
        panel = self.window.get_bottom_panel()
        panel.remove_item(self.notebook_bottompanel)
        for action, unused_handler in self.actions_visible_if_active:
            self.disconnect_object(action)
            action.set_visible(False)
        self.debug_active = False
        self.action_debug_toggle.set_active(False)
        
    def _guess_lang_shebang(self, path):
        try:
            with open(path, 'rb') as script:
                shebang = script.readline()
        except (OSError, IOError, TypeError):
            return None
        if not shebang.startswith(b'#!'):  return None
        if b'python3' in shebang:  return 'python3'
        if b'python' in shebang:  return 'python'
        return None
        
    def _guess_lang_path(self, path):
        lm = GtkSource.LanguageManager.get_default()
        lang = path and lm.guess_language(path, None)
        return lang and lang.get_id()
        
    def _guess_lang_document(self, doc):
        lang = doc and doc.get_language()
        return lang and lang.get_id()
        
    def start_debugging(self):
        '''Start debugger
        
        Show side panel and transfer data to the debugger (breakpoints).
        The debugged program stops at the first line.
        '''
        self.activate_debugger()
        self.unset_current_line()
        if self.debugger is not None:
            if self.debugger.running:
                return True
            self.debugger.deactivate()
            self.debugger = None
            
        path = self.entry_program.get_text()
        if not path:
            doc = self.window.get_active_document()
            location = doc and doc.get_location()
            path = location and location.get_path()
        else:
            doc = self.get_document_from_file(path)
        lang_id = self._guess_lang_shebang(path)
        if lang_id is None:
            lang_id = self._guess_lang_document(doc)
        if lang_id is None:
            lang_id = self._guess_lang_path(path)
            
        try:
            Debugger = debuggers[lang_id]
        except KeyError:
            self.terminal.feed('no debugger for lang {!r}\r\n'.format(lang_id).encode('utf-8'))
            return False
        self.debugger = Debugger(self.terminal)
        self.connectx(self.debugger, 'stack-changed', self.on_debugger_stack_changed)
        self.connectx(self.debugger, 'break-changed', self.on_debugger_break_changed)
        self.connectx(self.debugger, 'exited', self.on_debugger_exited)
        
        panel = self.window.get_side_panel()
        self.add_debugger_panel(panel, "DebuggerSidePanel", self.paned_sidepanel)
        
        args = self.get_arguments()
        workingdir = self.get_workingdir(path)
        self.debugger.start(path, args, workingdir)
        for cnum, snum, sourcemark, filename, lineno in self.breakpoints:
            doc = self.get_document_from_file(filename)
            if doc is None:
                self.debugger.bpadd(cnum, filename, lineno)
            else:
                titer = doc.get_iter_at_line(lineno - 1)
                if self.check_execline(doc, titer):
                    self.debugger.bpadd(cnum, filename, lineno)
                else:
                    doc.remove_source_marks(titer, titer, CATEGORY_BREAKPOINT)
        return False
        
    def end_debugging(self):
        '''Stop the debugger immediately
        
        Remove side panels as there is no useful data left to display.
        The debugger remains in the activated state.
        '''
        if self.debugger is None:
            return
        self.unset_current_line()
        self.set_frame_marks([])
        self.treeview_stack.set_model(None)
        self.treeview_namespace.set_model(None)
        self.debugger.deactivate()
        self.debugger = None
        
        panel = self.window.get_side_panel()
        panel.remove_item(self.paned_sidepanel)
        
    def get_arguments(self):
        args = self.entry_arguments.get_text()
        try:
            return shlex.split(args)
        except ValueError as e:
            #TODO: provide feedback in the gui
            print('{}: {}'.format(e, args))
            return []
        
    def get_workingdir(self, ref_filename):
        workingdir = self.entry_workingdir.get_text()
        if workingdir.startswith('/'):
            return workingdir
        dirname = os.path.dirname(ref_filename)
        dirname = os.path.join(dirname, workingdir)
        if os.path.isfile(dirname):
            dirname = os.path.dirname(dirname)
        return dirname
        
    def check_execline(self, doc, titer):
        lang = doc.get_language()
        try:
            Debugger = debuggers[lang.get_id()]
        except KeyError:
            return True
        end = titer.copy()
        if not end.ends_line():
            end.forward_to_line_end()
        line = titer.get_text(end)
        return Debugger.check_execline(line)
        
    def unset_current_line(self):
        if self.current_line and self.current_line[0]:
            self.current_line[0].delete_mark_by_name('debug-current-line')
            self.current_line = None
            
    def set_namespacestore(self, namespacestore):
        self.treeview_namespace.set_model(namespacestore.filter_new(Gtk.TreePath(self.frameno)))
        
    def set_doc_mark(self, doc, lineno, category):
        titer = doc.get_iter_at_line(lineno-1)
        return doc.create_source_mark(None, category, titer)
            
    def update_doc_mark(self, doc, sourcemark, lineno):
        if lineno > 0:
            titer = doc.get_iter_at_line(lineno - 1)
            doc.move_mark(sourcemark, titer)
        else:
            doc.delete_mark(sourcemark)
            
    def set_frame_marks(self, stack):
        # remove old source marks
        for filename in self.files_with_frame_marks:
            doc = self.get_document_from_file(filename)
            if doc is None:
                continue
            start, end = doc.get_bounds()
            doc.remove_source_marks(start, end, 'debug-frame')
        del self.files_with_frame_marks[:]
        # group by lineno
        filelines = {}
        for filename, lineno, unused_funcname, unused_tree in stack:
            filelines.setdefault(filename, []).append(lineno)
        # add new source marks
        for filename, lines in filelines.items():
            doc = self.get_document_from_file(filename)
            if doc is None:
                continue
            self.files_with_frame_marks.append(filename)
            for lineno in lines:
                self.set_doc_mark(doc, lineno, 'debug-frame')
                
    def get_keypath(self, tpath):
        def get_keys():
            while tpath.get_depth() > 0:
                yield namespacestore[tpath][0]
                tpath.up()
        namespacestore = self.treeview_namespace.get_model()
        keys = list(get_keys())
        keys.reverse()
        return keys
        
    def set_stack(self, stack):
        #       frameno, display filename, lineno, function, full filename
        stackstore = Gtk.ListStore(int, str, int, str, str)
        #       name, type, value, Pango.Weight
        namespacestore = Gtk.TreeStore(str, str, str, Pango.Weight)
        
        def add_rows_recursive(ptiter, children):
            for key, typestr, value, children in children:
                ctiter = namespacestore.append(ptiter, [key, typestr, value, Pango.Weight.NORMAL])
                add_rows_recursive(ctiter, children)
                
        for i, (filename, lineno, funcname, tree) in enumerate(reversed(stack)):
            if filename and filename[0] != '<':
                displayname = os.path.basename(filename)
            else:
                displayname = filename
                filename = None
            stackstore.append([i, displayname, lineno, funcname, filename])
            titer_frame = namespacestore.append(None, [str(i), '', '', Pango.Weight.NORMAL])
            for key, typestr, value, children in tree:
                titer = namespacestore.append(titer_frame, [key, typestr, value, Pango.Weight.BOLD])
                try:
                    add_rows_recursive(titer, children)
                except RuntimeError:
                    pass
        self.treeview_stack.set_model(stackstore)
        self.set_namespacestore(namespacestore)
        self.treeview_stack.set_cursor(Gtk.TreePath(self.frameno))
        
    def set_current_line(self, view, doc, filename, lineno):
        titer = doc.get_iter_at_line(lineno-1)
        view.scroll_to_iter(titer, 0.2, False, 0., 0.)
        self.unset_current_line()
        mark = doc.create_source_mark('debug-current-line', 'debug-current-line', titer)
        self.current_line = (doc, filename, lineno)
        
    def jump_to_line(self, filename, lineno):
        if not filename or not os.path.exists(filename):
            return
        location = Gio.File.new_for_path(filename)
        tab = self.window.get_tab_from_location(location)
        if tab is None:
            tab = self.window.create_tab_from_location(location, None, lineno, 0, False, True)
            view = tab.get_view()
        else:
            self.window.set_active_tab(tab)
            doc = tab.get_document()
            cur_iter = doc.get_iter_at_line(lineno-1)
            doc.place_cursor(cur_iter)
            view = tab.get_view()
            view.scroll_to_cursor()
        GLib.idle_add(view.grab_focus)
        
    def get_document_from_file(self, filename):
        location = Gio.File.new_for_path(filename)
        tab = self.window.get_tab_from_location(location)
        return tab and tab.get_document()
        
    def toggle_breakpoint(self, doc, titer):
        titer.set_line_offset(0)
        sourcemarks  = doc.get_source_marks_at_iter(titer, CATEGORY_BREAKPOINT)
        if sourcemarks:
            for sm in sourcemarks:
                snum = self.breakpoints.remove_by_sourcemark(sm)
                doc.delete_mark(sm)
                if snum is not None and self.debugger is not None and self.debugger.running:
                    self.debugger.bpclear(snum)
        elif self.check_execline(doc, titer):
            filename = doc.get_location().get_path()
            lineno = titer.get_line() + 1
            sourcemark = doc.create_source_mark(None, CATEGORY_BREAKPOINT, titer)
            cnum = self.breakpoints.add(sourcemark, filename, lineno)
            if self.debugger is not None and self.debugger.running:
                self.debugger.bpadd(cnum, filename, lineno)
            
    def on_tab_added(self, window, tab):
        doc = tab.get_document()
        view = tab.get_view()
        self.connectx(doc, 'loaded', self.on_document_loaded, view)
        self.connectx(view, 'line-mark-activated', self.on_view_line_mark_activated)
        
    def on_tab_removed(self, window, tab):
        doc = tab.get_document()
        if self.current_line and doc == self.current_line[0]:
            self.current_line = (None,) + self.current_line[1:]
        view = tab.get_view()
        self.disconnect_object(doc)
        self.disconnect_object(view)
        
    def on_document_loaded(self, doc, unused_error, view):
        if self.current_line and self.current_line[0] is None:
            if doc.get_location().get_path() == self.current_line[1]:
                self.set_current_line(view, doc, *self.current_line[1:])
        stackstore = self.treeview_stack.get_model()
        filename = doc.get_location().get_path()
        if stackstore is not None:
            for row in stackstore:
                if filename == row[4]:
                    self.set_doc_mark(doc, row[2], 'debug-frame')
        for row, lineno in self.breakpoints.lines_in_file(filename):
            sourcemark = self.set_doc_mark(doc, lineno, CATEGORY_BREAKPOINT)
            self.breakpoints.update_sourcemark(row, sourcemark)
        
    def on_view_line_mark_activated(self, view, titer, event):
        if event.type != Gdk.EventType.BUTTON_PRESS:
            return
        # event.button is the real event
        if event.button.button != 1 or event.button.state & self.modifier_mask:
            return
        doc = view.get_buffer()
        self.toggle_breakpoint(doc, titer)
        
    def on_action_debug_toggle(self, action):
        if self.debug_active:
            self.deactivate_debugger()
        else:
            self.activate_debugger()
            
    def on_action_step_activate(self, action):
        if self.start_debugging():
            self.debugger.step()
        
    def on_action_next_activate(self, action):
        if self.start_debugging():
            self.debugger.next()
        
    def on_action_return_activate(self, action):
        if self.start_debugging():
            self.debugger.goreturn()
        
    def on_action_run_to_activate(self, action):
        doc = self.window.get_active_document()
        titer = doc.get_iter_at_mark(doc.get_insert())
        titer.set_line_offset(0)
        if self.check_execline(doc, titer):
            filename = doc.get_location().get_path()
            lineno = titer.get_line() + 1
            if self.start_debugging():
                self.debugger.run_to(filename, lineno)
                
    def on_action_jump_to_activate(self, action):
        doc = self.window.get_active_document()
        titer = doc.get_iter_at_mark(doc.get_insert())
        titer.set_line_offset(0)
        if self.check_execline(doc, titer):
            filename = doc.get_location().get_path()
            lineno = titer.get_line() + 1
            if self.start_debugging():
                self.debugger.jump_to(filename, lineno)
                
    def on_action_cont_activate(self, action):
        if self.start_debugging():
            self.debugger.cont(nopause=False)
        
    def on_action_cont_nopause_activate(self, action):
        if self.start_debugging():
            self.debugger.cont(nopause=True)
            
    def on_action_pause_activate(self, action):
        if self.debugger and self.debugger.running:
            self.debugger.pause()
            
    def on_action_break_activate(self, action):
        doc = self.window.get_active_document()
        titer = doc.get_iter_at_mark(doc.get_insert())
        self.toggle_breakpoint(doc, titer)
            
    def on_action_quit_activate(self, action):
        self.end_debugging()
        
    def on_debugger_stack_changed(self, debugger, stack, keepframe):
        if not keepframe:
            self.frameno = 0
        self.set_frame_marks(stack)
        self.set_stack(stack)
        filename, lineno, unused_funcname, unused_tree = stack[-1]
        if filename and filename[0] != '<':
            # show current line
            location = Gio.File.new_for_path(filename)
            tab = self.window.get_tab_from_location(location)
            if tab is None:
                self.unset_current_line()
                self.current_line = (None, filename, lineno)
                tab = self.window.create_tab_from_location(location, None, lineno, 0, False, True)
            else:
                self.window.set_active_tab(tab)
                self.set_current_line(tab.get_view(), tab.get_document(), filename, lineno)
        
    def on_debugger_break_changed(self, debugger, breakpoint):
        sourcemark, filename, lineno1, lineno2 = self.breakpoints.update(breakpoint)
        if lineno1 == lineno2:
            return
        doc = self.get_document_from_file(filename)
        if doc is None:
            return
        self.update_doc_mark(doc, sourcemark, lineno2)
            
    def on_debugger_exited(self, debugger):
        self.end_debugging()
        
    def on_treeview_stack_cursor_changed(self, treeview):
        tpath = self.treeview_stack.get_cursor()[0]
        if tpath is None:
            return
        stackstore = self.treeview_stack.get_model()
        self.frameno = stackstore[tpath][0]
        
        namespacestore = self.treeview_namespace.get_model().get_model()
        self.set_namespacestore(namespacestore)
        self.treeview_namespace.expand_all()
        def keypath_to_tpath(rows, keypath, tpath):
            if not keypath:
                return tpath
            key, *keypath = keypath
            for row in rows:
                if row[0] == key:
                    return keypath_to_tpath(row.iterchildren(), keypath, row.path)
            return tpath
        tpath = keypath_to_tpath(self.treeview_namespace.get_model(), self.keypath, None)
        if tpath is not None:
            self.no_update_keypath = True
            self.treeview_namespace.set_cursor(tpath)
        
    def on_treeview_stack_row_activated(self, treeview, tpath, unused_column):
        stackstore = self.treeview_stack.get_model()
        filename = stackstore[tpath][4]
        lineno = stackstore[tpath][2]
        self.jump_to_line(filename, lineno)
        
    def on_treeview_namespace_cursor_changed(self, treeview):
        if self.no_update_keypath:
            self.no_update_keypath = False
            return
        tpath = treeview.get_cursor()[0]
        if tpath is not None:
            self.keypath = self.get_keypath(tpath)
            
    def on_treeview_namespace_row_activated(self, treeview, tpath, unused_column):
        if not (self.debugger and self.debugger.running):
            return
        if tpath.get_depth() <= 0:
            return
        keys = self.get_keypath(tpath)
        self.debugger.inspect(keys)
        
    def on_notebook_bottompanel_switch_page(self, notebook, widget, page_num):
        if widget == self.grid_parameters:
            GLib.idle_add(self.entry_arguments.grab_focus)
            
    def on_treeview_breakpoints_row_activated(self, unused_treeview, tpath, unused_column):
        tpath = self.breakpoints.child_path(tpath)
        filename, lineno = self.breakpoints.fileinfo(tpath)
        self.jump_to_line(filename, lineno)
        
    def on_treeview_breakpoints_key_release_event(self, treeview, event):
        if event.keyval != Gdk.KEY_Delete or event.state & self.modifier_mask:
            return
        tpath = treeview.get_cursor()[0]
        if tpath is None:
            return
        tpath = self.breakpoints.child_path(tpath)
        snum, sourcemark = self.breakpoints.identifiers(tpath)
        self.breakpoints.remove(self.breakpoints.get_iter(tpath))
        doc = sourcemark and sourcemark.get_buffer()
        if doc is not None:
            doc.delete_mark(sourcemark)
        if self.debugger is not None and self.debugger.running:
            self.debugger.bpclear(snum)
            
    def on_entry_arguments_activate(self, unused_entry):
        if self.debugger and self.debugger.running:
            self.debugger.chargs(self.get_arguments())
            
    def on_entry_workingdir_activate(self, unused_entry):
        if self.debugger and self.debugger.running:
            workingdir = self.get_workingdir(self.debugger.filename)
            self.debugger.chdir(workingdir)
                
    def on_button_set_program_activate(self, unused):
        doc = self.window.get_active_document()
        location = doc and doc.get_location()
        path = location and location.get_path()
        path and self.entry_program.set_text(path)
        
    def on_button_set_workingdir_activate(self, unused):
        doc = self.window.get_active_document()
        location = doc and doc.get_location()
        path = location and location.get_path()
        path = path and os.path.dirname(path)
        path and self.entry_workingdir.set_text(path)
        
        

